package ai;

import attacktrees.*;
import myutil.TraceManager;
import org.json.JSONArray;
import org.json.JSONObject;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;

/**
 * Class AIUseCaseDiagram
 * <p>
 * Creation: 19/03/2024
 *
 * @author Alan Birchler De Allende
 */
public class AIAttackPatternTree3 extends AIInteract {
    private static final String KNOWLEDGE_ON_JSON_FOR_ROOT = "When you are asked to identify a motive, " +
            "return the motive formatted as JSON like so: " +
            "{\"rootattack\": {\"name\":  \"NameOfMotive\", \"description\": \"" +
            "The description of the motive.\"}} " +
            "# Respect: All words in \"name\" must be conjoined together. " +
            "# Respect: There must be no more than forty characters in \"name\". " +
            "# Respect: For each word in \"name\", its first letter must be capitalized. " +
            "# Respect: All words in \"description\" must be separated with spaces.";

    private static final String KNOWLEDGE_ON_JSON_FOR_ATTACK_SCEN = "When you are asked to identify all the possible " +
            "attack scenarios that an attacker would perform to successfully achieve an " +
            "attack, return them as a JSON specification formatted as follows: " +
            "{\"attack\": \"NameOfAttack\", \"attackscenarios\": [{\"name\":  \"NameOfAttackScenario\", \"description\": \"" +
            "The description of the attack scenario and how it brings an attacker closer to the attack.\", " +
            "\"operator\": \"OR or AND\"} ...]} " +
            "# Respect: All words in \"name\" must be conjoined together. " +
            "# Respect: \"attack\" must be the name of the provided attack. " +
            "# Respect: There must be no more than forty characters in \"name\". " +
            "# Respect: For each word in \"name\", its first letter must be capitalized. " +
            "# Respect: Include what the attack scenario is and how it is used by an attacker for getting closer " +
            "to the attack in \"description\". " +
            "# Respect: All words in \"description\" must be separated with spaces. " +
            "# Respect: \"operator\" must be only \"AND\" or \"OR\". Use \"AND\" to denote that an attacker must " +
            "complete all of the attack scenarios with an AND operator simultaneously to get " +
            "closer to the attack. Use \"OR\" to denote that an attacker only needs to complete one of the attack " +
            "scenarios amongst all of the attack scenarios with an OR operator to get closer to the attack. " +
            "# Respect: The number of attack scenarios with the AND operator should be either zero, or greater than one. " +
            "# Respect: The number of attack scenarios with the OR operator should be either zero, or greater than one. " +
            "# Respect: If there are no attack scenarios that are able to be identified, have \"attackscenarios\" be " +
            "an empty array.";

    private static final String KNOWLEDGE_ON_JSON_FOR_ATT_STEPS = "When you are asked to identify all of the steps that " +
            "that an attacker needs to complete to successfully perform an attack scenario, " +
            "return them as a JSON specification formatted as follows: " +
            "{\"attackscenario\": \"NameOfAttackScenario\", \"attacksteps\": [{\"name\":  \"NameOfAttackStep\", \"description\": \"" +
            "The description of the attack step and how it is needed to achieve the attack scenario.\"} ...]} " +
            "# Respect: All words in \"name\" must be conjoined together. " +
            "# Respect: \"attackscenario\" must be the name of the provided attack scenario. " +
            "# Respect: There must be no more than forty characters in \"name\". " +
            "# Respect: For each word in \"name\", its first letter must be capitalized. " +
            "# Respect: Include what the attack step is and how it needs to be completed by an attacker for " +
            "the attack scenario in \"description\". " +
            "# Respect: All words in \"description\" must be separated with spaces. " +
            "# Respect: The elements of \"attacksteps\" need to be ordered sequentially. That is, the first indexed element " +
            "in \"attacksteps\" is the step that an attacker needs to complete first while the last indexed element is the step " +
            "that an attacker needs to complete last. " +
            "# Respect: There must be at least two attack steps in \"attacksteps\".";

//    private static final String KNOWLEDGE_ON_JSON_FOR_MITIGATIONS = "When you are asked to identify the mitigations " +
//            "that prevent an attacker from performing an attack node, " +
//            "return them as a JSON specification formatted as follows: " +
//            "{\"mitigations\": [{\"name\":  \"NameOfMitigation\", \"description\": \"" +
//            "The description of the mitigation and how it prevents an attacker from completing " +
//            "its associated attack node.\"} ...]} " +
//            "# Respect: All words in the \"name\" of each mitigation must be conjoined together. " +
//            "# Respect: There must be no more than forty characters in the \"name\" of each mitigation. " +
//            "# Respect: For each word in each mitigation's \"name\", its first letter must be capitalized. " +
//            "# Respect: Include what the mitigation is and how it is used to prevent an attacker from " +
//            "completing its associated attack node in the same \"description\" key. " +
//            "# Respect: If there are no mitigations that can be applied to the provided attack nodes, return the " +
//            "following JSON: {\"mitigations\": \"No mitigations were identified.\"} " +
//            "# Respect: All words in the \"description\" key must be separated with spaces. ";
//
//    private static final String KNOWLEDGE_ON_JSON_FOR_MITI_PAIRS = "When you are asked to identify the " +
//            "mitigation that can be applied to an attack node to prevent " +
//            "an attacker from performing the attack node, " +
//            "return them as a JSON specification formatted as follows: " +
//            "{\"mitigationpairings\": [{\"mitigation\":  \"NameOfMitigation\", \"attacknode\": \"NameOfAttackNode\"} ...]} " +
//            "# Respect: The value of \"mitigation\" should be only one of the names of the given mitigations. " +
//            "# Respect: The value of \"attacknode\" should be only one of the names of the given attack nodes. " +
//            "# Respect: The value of \"attacknode\" should not be the name of the root attack.";

    private static final String[] KNOWLEDGE_STAGES = {
            KNOWLEDGE_ON_JSON_FOR_ROOT,
            KNOWLEDGE_ON_JSON_FOR_ATTACK_SCEN,
            KNOWLEDGE_ON_JSON_FOR_ATT_STEPS
//            KNOWLEDGE_ON_JSON_FOR_MITIGATIONS,
//            KNOWLEDGE_ON_JSON_FOR_MITI_PAIRS
    };

    private final String[] QUESTION_IDENTIFY_ATD = {"From the provided system specification " +
            "and using the specified JSON format, identify a motive as to why an attacker would " +
            "want to exploit the system. If provided with a list " +
            "of attack patterns, use at least one attack pattern from this list to identify a possible motive. " +
            "Do respect the JSON format, and provide only JSON (no explanation before or after). \n",

            "Using the specified JSON format, " +
                    "identify all of the attack scenarios that an attacker would perform, " +
                    "either simultaneously or separately, to achieve the provided attack. If provided with a " +
                    "system specification, make sure to associate the specification with the attack scenarios. " +
                    "In addition, if provided with a list of attack patterns, use at least one attack pattern " +
                    "from this list to identify the attack scenarios. Do respect the JSON format, and provide " +
                    "only JSON (no explanation before or after).\n",

            "Identify all of the attack steps that an attacker needs to conduct to " +
                    "achieve the provided attack scenario. Do respect the JSON format, and " +
                    "provide only JSON (no explanation before or after).\n",

//            "From the provided system specification, attack pattern, and attack nodes " +
//                    "and using the specified JSON format, " +
//                    "identify possible mitigations, if there are any, that could prevent an attacker from completing " +
//                    "an attack node. Do respect the JSON format, and provide only JSON (no explanation before or after).\n",
//
//            "From the provided system specification, attack pattern, attack nodes, and mitigations " +
//                    "and using the specified JSON format, " +
//                    "identify what mitigation can be applied to which attack node such that the mitigation prevents " +
//                    "an attacker from performing the attack node. " +
//                    "Do respect the JSON format, and provide only JSON (no explanation before or after).\n"
    };

    private String rootAttack;
    private ArrayList<JSONArray> attackScenarios;
    private ArrayList<JSONArray> attackSteps;
    private AttackTree at;
    private int maxNumAts;
    private int maxLevels;
    private int maxRetries;
    private final ArrayList<AttackTree> atList;

    public AIAttackPatternTree3(AIChatData _chatData) {
        super(_chatData);
        atList = new ArrayList<>();
        maxRetries = 5;
        maxLevels = 1;
        maxNumAts = 1;
    }

    public void setMaxRetries (int maxRetries) {
        this.maxRetries = maxRetries;
    }

    public void setMaxLevels (int maxLevels) {
        this.maxLevels = maxLevels;
    }

    public void setMaxNumAts (int maxNumAts) {
        this.maxNumAts = maxNumAts;
    }

    public AttackTree getATDiagram() {
        return at;
    }

    public Object applyAnswer(Object input) {
        return atList;
    }

    private void initKnowledge() {
        chatData.aiinterface.clearKnowledge();
    }

    private void makeKnowledge(int stage, String _spec, String[] _attackPatterns, String previousRootAtts) {
        String [] know = KNOWLEDGE_STAGES[stage].split("#");
        for(String s: know) {
            TraceManager.addDev("\nKnowledge added: " + s);
            chatData.aiinterface.addKnowledge(s, "ok");
        }

        if (_spec != null) {
            TraceManager.addDev("\nKnowledge added: " + _spec);
            chatData.aiinterface.addKnowledge("The system specification is: " + _spec, "ok");
        }

        if (_attackPatterns != null && _attackPatterns.length > 0) {
            StringBuilder builder = new StringBuilder();

            for (int i = 0; i <= _attackPatterns.length - 1; i++) {
                String attackPatternCleaned = _attackPatterns[i].
                        replaceAll("Confidence score: [0-9]+%", "").
                        replace("\n", " ").
                        trim();
                builder.append(attackPatternCleaned);

                if (i < _attackPatterns.length - 1) {
                    builder.append(" | ");
                }
            }

            TraceManager.addDev("\nKnowledge added: " + builder);
            chatData.aiinterface.addKnowledge("The attack patterns are: " + builder, "ok");
        }

        if (previousRootAtts != null && !previousRootAtts.isEmpty()) {
            TraceManager.addDev("\nKnowledge added: " + previousRootAtts);
            chatData.aiinterface.addKnowledge(
                    "Identify a motive other than the motives in the following JSONs: " + previousRootAtts,
                    "ok");

        }
    }

    public void internalRequest() {
        String previousRootAttacks = "";

        for (int z = 0; z < maxNumAts; z++) {
            at = new AttackTree("", null);
            rootAttack = "";
            attackSteps = new ArrayList<>();
            int stage = 0;
            int numLevels = 0;
            int numTries = 0;
            boolean done = false;

            String lastQuestion = chatData.lastQuestion.trim();
            String[] data = lastQuestion.split("\n\n");
            String systemSpec = data[0];
            String[] attackPatterns = Arrays.copyOfRange(data, 1, data.length);

            while (!done) {
                if (stage % 2 == 1) {
                    attackScenarios = new ArrayList<>();
                }
                else if (stage % 2 == 0 && stage != 0) {
                    attackSteps = new ArrayList<>();
                }

                if (stage < 2) {
                    numTries = makeRequest(stage, systemSpec, attackPatterns, null, previousRootAttacks);
                }
                else if (stage % 2 == 0) {
                    for (int i = 0; i < attackScenarios.size(); i++) {
                        JSONArray attackScenarioArray = attackScenarios.get(i);

                        for (int k = 0; k < attackScenarioArray.length(); k++) {
                            String attackScenarioString = attackScenarioArray.get(k).toString();
                            numTries = makeRequest(stage, systemSpec, attackPatterns, attackScenarioString, null);
                        }
                    }
                }
                else {
                    for (int i = 0; i < attackSteps.size(); i++) {
                        JSONArray attackStepArray = attackSteps.get(i);

                        for (int k = 0; k < attackStepArray.length(); k++) {
                            String attackStepString = attackStepArray.get(k).toString();
                            numTries = makeRequest(stage, systemSpec, attackPatterns, attackStepString, null);
                        }
                    }
                }

                if (stage % 2 == 0 && stage != 0) {
                    numLevels++;
                }

                if (numTries >= maxRetries || numLevels >= maxLevels) {
                    done = true;
                }

                stage++;
                waitIfConditionTrue(!done);
            }

            TraceManager.addDev("Reached end of AIAttackPatternTree internal request.");
            atList.add(at);
            previousRootAttacks = previousRootAttacks + rootAttack + " ";
        }
    }

    private int makeRequest(int stage, String systemSpec, String[] attackPatterns, String attStSc, String previousRootAtts) {
        int numRetries = 0;
        boolean done = false;
        String json = "";
        String questionT = "";

        if (stage == 0) {
            questionT = QUESTION_IDENTIFY_ATD[stage];
        }
        else if (stage % 2 == 1) {
            questionT = QUESTION_IDENTIFY_ATD[1];
        }
        else if (stage % 2 == 0) {
            questionT = QUESTION_IDENTIFY_ATD[2];
        }

        if (stage == 1) {
            questionT += "\nThe attack data is in the following JSON:\n" + rootAttack.trim() + "\n";
        }
        else if (stage % 2 == 0 && stage != 0) {
            questionT += "\nThe attack scenario data is in the following JSON:\n" + attStSc.trim() + "\n";
        }
        else if (stage % 2 == 1) {
            questionT += "\nThe attack data is in the following JSON:\n" + attStSc.trim() + "\n";
        }

        while (numRetries < maxRetries && !done) {
            initKnowledge();

            if (stage == 0) {
                makeKnowledge(stage, systemSpec, attackPatterns, previousRootAtts);
            }
            else if (stage == 1) {
                makeKnowledge(stage, systemSpec, attackPatterns, null);
            }
            else if (stage % 2 == 0){
                makeKnowledge(2, null, null, null);
            }
            else {
                makeKnowledge(1, null, null, null);
            }

            boolean ok = makeQuestion(questionT);

            if (!ok) {
                TraceManager.addDev("Make question failed");
                return maxRetries;
            }

            ArrayList<String> errors = new ArrayList<>();

            try {
                TraceManager.addDev("\n\nMaking specification from " + chatData.lastAnswer + "\n\n");
                json = extractJSON();

                if (stage == 0) {
                    rootAttack = checkRootAttack(json, errors);
                    TraceManager.addDev("Identified consquence - " + rootAttack);

                    if (rootAttack.isEmpty()) {
                        errors.add("You must provide a motive that an attacker would " +
                                "have as to why they would want to exploit the provided system specification. " +
                                "Do respect the JSON " +
                                "format, and provide only JSON (no explanation before or after).");
                    }
                }
                else if (stage % 2 == 1) {
                    JSONArray attackScenarioJSON;

                    if (stage == 1) {
                        attackScenarioJSON = checkAttackScenarios(json, rootAttack, true, errors, stage);

                    }
                    else {
                        attackScenarioJSON = checkAttackScenarios(json, attStSc, false, errors, stage);
                    }

                    TraceManager.addDev("Identified attack scenarios: " + attackScenarioJSON);

                    if (attackScenarioJSON.isEmpty() && !errors.isEmpty()) {
                        errors.add("You must provide all possible attack scenarios showing how an " +
                                "attacker uses them to achieve the provided attack. " +
                                "Do respect the JSON format, and " +
                                "provide only JSON (no explanation before or after).");
                    }

                    if (!attackScenarioJSON.isEmpty()) {
                        attackScenarios.add(attackScenarioJSON);
                    }
                }
                else if (stage % 2 == 0) {
                    JSONArray attackStepJSON = checkAttackSteps(json, attStSc, errors);
                    TraceManager.addDev("Identified attack steps: " + attackStepJSON);

                    if (attackStepJSON.isEmpty()) {
                        errors.add("You must provide all of the attack steps needed for an attacker" +
                                "to achieve the provided attack scenario. " +
                                "Do respect the JSON format, and " +
                                "provide only JSON (no explanation before or after).");
                    }
                    else {
                        attackSteps.add(attackStepJSON);
                    }
                }
            } catch (org.json.JSONException e) {
                TraceManager.addDev("Invalid JSON spec: " + extractJSON() + " because " + e.getMessage() + ": INJECTING ERROR");
                errors = new ArrayList<>();
                errors.add("There is an error in your JSON: " + e.getMessage() + ". Probably the JSON spec was incomplete. Do correct it. I need " +
                        "the full specification at once.");
            }

            if (!errors.isEmpty()) {
                questionT = "Your answer was as follows: " + json + "\n\nYet, it was not correct because of the following errors:";
                // Updating knowledge
                for (String s : errors) {
                    questionT += "\n- " + s;
                }

                numRetries++;
            } else {
                done = true;
            }
        }

        return numRetries;
    }

    private String checkRootAttack(String _spec, Collection<String> _errors) throws org.json.JSONException {
        if (_spec == null) {
            _errors.add("No \"rootattack\" object in json");

            return "";
        }

        int indexStart = _spec.indexOf('{');
        int indexStop = _spec.lastIndexOf('}');

        if ((indexStart == -1) || (indexStop == -1) || (indexStart > indexStop)) {
            _errors.add("Invalid JSON object (start or stop)");

            return "";
        }

        String json = _spec.substring(indexStart, indexStop + 1);

        JSONObject mainObject = new JSONObject(json);
        JSONObject rootAttackJSON = mainObject.getJSONObject("rootattack");

        if (rootAttackJSON == null) {
            TraceManager.addDev("No \"rootattack\" object in json");
            _errors.add("No \"rootattack\" object in json");

            return "";
        }
        else {
            _errors.addAll(checkAttackStepSyntax(rootAttackJSON));
        }

        String name = rootAttackJSON.getString("name");

        for (AttackTree attackTree : atList) {
            for (Attack attack : attackTree.getAttacks()) {
                if (attack.isRoot() && attack.getName().equals(name)) {
                    TraceManager.addDev("Duplicate motive identified");
                    _errors.add("The motive that you provided has already been identified. " +
                            "Please provide another motive.");

                    return "";
                }
            }
        }

        if (_errors.isEmpty()) {
            Attack rootAttack = new Attack("", null);
            rootAttack.setRoot(true);
            rootAttack.setName(rootAttackJSON.getString("name"));
            rootAttack.setDescription(rootAttackJSON.getString("description"));
            at.addAttack(rootAttack);

            return _spec;
        }
        else {
            return "";
        }
    }

    private JSONArray checkAttackScenarios(String _spec, String attack, boolean isRoot,
                                           Collection<String> _errors, int stage) throws org.json.JSONException {
        String attackName;

        if (isRoot) {
            attackName = extractAttackName(extractRootAttack(attack));
        }
        else {
            attackName = extractAttackName(attack);
        }

        if (_spec == null) {
            _errors.add("No \"attack\" object in json");

            return new JSONArray();
        }

        int indexStart = _spec.indexOf('{');
        int indexStop = _spec.lastIndexOf('}');

        if ((indexStart == -1) || (indexStop == -1) || (indexStart > indexStop)) {
            _errors.add("Invalid JSON object (start or stop)");

            return new JSONArray();
        }

        String json = _spec.substring(indexStart, indexStop + 1);

        JSONObject mainObject = new JSONObject(json);
        String jsonAttackName = mainObject.getString("attack");
        JSONArray attackScenariosJSON = mainObject.getJSONArray("attackscenarios");

        if (jsonAttackName == null) {
            TraceManager.addDev("No \"attack\" in json");
            _errors.add("No \"attack\" in json");

            return new JSONArray();
        }
        else if (!attackName.equals(jsonAttackName)) {
            TraceManager.addDev("The value of \"attack\" should be set to " + attackName);
            _errors.add("The value of \"attack\" should be set to " + attackName);

            return new JSONArray();
        }
        else if (attackScenariosJSON == null) {
            TraceManager.addDev("No \"attackscenarios\" array in json");
            _errors.add("No \"attackscenarios\" array in json");

            return new JSONArray();
        }
        else if (attackScenariosJSON.isEmpty()) {
            if (stage == 1) {
                TraceManager.addDev("\"attackscenarios\" json array is empty.");
                _errors.add("\"attackscenarios\" json array is empty. Please make sure to " +
                        "include at least two attack scenarios.");
            }

            return new JSONArray();
        }

        String redundantAttScen = hasOnlyOneAttScen(attackScenariosJSON);

        if (redundantAttScen != null) {
            TraceManager.addDev(redundantAttScen + "is a sole OR or AND attack scenario");

            if (stage == 1) {
                _errors.add(redundantAttScen + " is a sole OR or AND attack scenario. " +
                        "Please provide additional attack scenarios.");
            }
            else {
                _errors.add(redundantAttScen + " is a sole OR or AND attack scenario. " +
                        "Please remove it from \"attackscenarios\".");
            }

            return new JSONArray();
        }
        else {
            int i = 0;

            while (i < attackScenariosJSON.length() && _errors.isEmpty())
            {
                _errors.addAll(checkAttackScenSyntax(attackScenariosJSON.getJSONObject(i)));
                i++;
            }
        }

        if (_errors.isEmpty()) {
            createAttackScenarios(attackScenariosJSON, jsonAttackName);

            return attackScenariosJSON;
        }
        else {
            return new JSONArray();
        }
    }

    private String hasOnlyOneAttScen(Iterable<Object> jsonArray) {
        ArrayList<String> orCount = new ArrayList<>();
        ArrayList<String> andCount = new ArrayList<>();

        for (Object object : jsonArray) {
            JSONObject json = (JSONObject) object;
            String operator = json.getString("operator");
            String name = json.getString("name");

            if (operator.equals("OR")) {
                orCount.add(name);
            }
            else if (operator.equals("AND")) {
                andCount.add(name);
            }
        }

        if (orCount.size() == 1) {
            return orCount.get(0);
        }
        else if (andCount.size() == 1) {
            return andCount.get(0);
        }
        else {
            return null;
        }
    }

    private ArrayList<String> checkAttackScenSyntax(JSONObject attack) {
        ArrayList<String> errors = new ArrayList<>();

        if (attack.length() != 3) {
            errors.add("The attack object should only contain \"name\", \"description\", " +
                    "and \"operator\"");

            return errors;
        }

        String name = attack.getString("name");
        String description = attack.getString("description");
        String operator = attack.getString("operator");

        if (name == null) {
            errors.add("Attack scenario has no name");
            return errors;
        }

        Attack alreadyPresentAtt = findAttack(name);

        if (alreadyPresentAtt != null) {
            errors.add(name + " is already the name of an attack scenario or step that you created. " +
                    "Please remove it from \"attackscenarios\".");
        }
        else if (!name.matches("[a-zA-Z0-9]+")) {
            errors.add(name + " must only contain alphanumeric characters");
        }
        else if (name.length() > 40) {
            errors.add(name + " must only contain forty characters max");
        }
        else if (description == null) {
            errors.add("Attack has no description");
            return errors;
        }
        else if (!description.contains(" ")) {
            errors.add("The words in \"description\" must be separated with spaces");
        }
        else if (!operator.equals("OR") && !operator.equals("AND")) {
            errors.add("The value of \"operator\" must be either OR or AND");
        }

        return errors;
    }

    private void createAttackScenarios (JSONArray attackScenarios, String parentAttackName) {
        AttackNode andConnection = null;
        AttackNode orConnection = null;
        Attack parentAttack = findAttack(parentAttackName);

        for (int i = 0; i < attackScenarios.length(); i++) {
            JSONObject attackScen = attackScenarios.getJSONObject(i);
            Attack attack = new Attack("", null);
            attack.setName(attackScen.getString("name"));
            attack.setDescription(attackScen.getString("description"));
            at.addAttack(attack);

            String operator = attackScen.getString("operator");

            if (operator.equals("OR")) {
                if (orConnection == null) {
                    orConnection = new ORNode("", null);
                    orConnection.setResultingAttack(parentAttack);
                    at.addNode(orConnection);
                }

                orConnection.addInputAttack(attack, i + 1);
            }
            else {
                if (andConnection == null) {
                    andConnection = new ANDNode("", null);
                    andConnection.setResultingAttack(parentAttack);
                    at.addNode(andConnection);
                }

                andConnection.addInputAttack(attack, i + 1);
            }
        }
    }

    private JSONArray checkAttackSteps(String _spec, String attack, Collection<String> _errors)
            throws org.json.JSONException {
        String attackName = extractAttackName(attack);

        if (_spec == null) {
            _errors.add("No \"attackscenario\" object in json");

            return new JSONArray();
        }

        int indexStart = _spec.indexOf('{');
        int indexStop = _spec.lastIndexOf('}');

        if ((indexStart == -1) || (indexStop == -1) || (indexStart > indexStop)) {
            _errors.add("Invalid JSON object (start or stop)");

            return new JSONArray();
        }

        String json = _spec.substring(indexStart, indexStop + 1);

        JSONObject mainObject = new JSONObject(json);
        String jsonAttScenName = mainObject.getString("attackscenario");
        JSONArray attackStepsJSON = mainObject.getJSONArray("attacksteps");

        if (jsonAttScenName == null) {
            TraceManager.addDev("No \"attackscenario\" in json");
            _errors.add("No \"attackscenario\" in json");

            return new JSONArray();
        }
        else if (!attackName.equals(jsonAttScenName)) {
            TraceManager.addDev("The value of \"attackscenario\" should be set to " + attackName);
            _errors.add("The value of \"attackscenario\" should be set to " + attackName);

            return new JSONArray();
        }
        else if (attackStepsJSON == null) {
            TraceManager.addDev("No \"attacksteps\" array in json");
            _errors.add("No \"attacksteps\" array in json");

            return new JSONArray();
        }
        else if (attackStepsJSON.length() <= 1) {
            TraceManager.addDev("Only one provided attack step");
            _errors.add("Please provide at least two attack steps in \"attacksteps\"");

            return new JSONArray();
        }
        else {
            int i = 0;

            while (i < attackStepsJSON.length() && _errors.isEmpty())
            {
                _errors.addAll(checkAttackStepSyntax(attackStepsJSON.getJSONObject(i)));
                i++;
            }
        }

        if (_errors.isEmpty()) {
            createAttackSteps(attackStepsJSON, jsonAttScenName);

            return attackStepsJSON;
        }
        else {
            return new JSONArray();
        }
    }

    private ArrayList<String> checkAttackStepSyntax(JSONObject attackStep) {
        ArrayList<String> errors = new ArrayList<>();

        if (attackStep.length() != 2) {
            errors.add("The JSON object should only contain \"name\" and \"description\"");

            return errors;
        }

        String name = attackStep.getString("name");
        String description = attackStep.getString("description");

        if (name == null) {
            errors.add("Attack has no name");
            return errors;
        }

        Attack alreadyPresentAtt = findAttack(name);

        if (alreadyPresentAtt != null) {
            errors.add(name + " is already the name of an attack scenario or step that you created. " +
                    "Please remove it from the list.");
        }
        else if (!name.matches("[a-zA-Z0-9]+")) {
            errors.add(name + " must only contain alphanumeric characters");
        }
        else if (name.length() > 40) {
            errors.add(name + " must only contain forty characters max");
        }
        else if (description == null) {
            errors.add("The JSON object has no description");
            return errors;
        }
        else if (!description.contains(" ")) {
            errors.add("The words in \"description\" must be separated with spaces");
        }

        return errors;
    }

    private void createAttackSteps(JSONArray attackSteps, String parentAttackName) {
        AttackNode seqConnection = new SequenceNode("", null);
        Attack parentAttack = findAttack(parentAttackName);
        seqConnection.setResultingAttack(parentAttack);
        at.addNode(seqConnection);

        for (int i = 0; i < attackSteps.length(); i++) {
            JSONObject attackStep = attackSteps.getJSONObject(i);
            Attack attack = new Attack("", null);
            attack.setName(attackStep.getString("name"));
            attack.setDescription(attackStep.getString("description"));
            seqConnection.addInputAttack(attack, i + 1);

            at.addAttack(attack);
        }
    }

    private Attack findAttack(String attackNodeName) {
        String attackNNLower = attackNodeName.toLowerCase();
        ArrayList<Attack> attackNodeList = at.getAttacks();
        Attack attack;
        int i = 0;

        while (i < attackNodeList.size()) {
            attack = attackNodeList.get(i);

            if (attackNNLower.equals(attack.getName().toLowerCase())) {
                return attack;
            }

            i++;
        }

        return null;
    }

    private String extractRootAttack(String _spec) throws org.json.JSONException {
        if (_spec == null) {
            return "";
        }

        int indexStart = _spec.indexOf('{');
        int indexStop = _spec.lastIndexOf('}');

        if ((indexStart == -1) || (indexStop == -1) || (indexStart > indexStop)) {
            return "";
        }

        String json = _spec.substring(indexStart, indexStop + 1);
        JSONObject mainObject = new JSONObject(json);
        JSONObject rootAttack = mainObject.getJSONObject("rootattack");

        if (rootAttack != null) {
            return rootAttack.toString();
        }

        return "";
    }

    private String extractAttackName(String _spec) throws org.json.JSONException {
        if (_spec == null) {
            return "";
        }

        int indexStart = _spec.indexOf('{');
        int indexStop = _spec.lastIndexOf('}');

        if ((indexStart == -1) || (indexStop == -1) || (indexStart > indexStop)) {
            return "";
        }

        String json = _spec.substring(indexStart, indexStop + 1);
        JSONObject mainObject = new JSONObject(json);
        String attackName = mainObject.getString("name");

        if (attackName != null) {
            return attackName;
        }

        return "";
    }
}
