package ai;

import attacktrees.*;
import myutil.TraceManager;

import org.json.JSONArray;
import org.json.JSONObject;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

/**
 * Class AIUseCaseDiagram
 * <p>
 * Creation: 19/03/2024
 *
 * @author Alan Birchler De Allende
 */
public class AIAttackPatternTree1 extends AIInteract {
    private static final String KNOWLEDGE_ON_JSON_FOR_ROOT = "When you are asked to identify the root attack of " +
            "an attack pattern and how it can be used to exploit a provided system specification, " +
            "return the root attack formatted as JSON like so: " +
            "{\"rootattack\": {\"name\":  \"NameOfRootAttack\", \"description\": \"" +
            "The description of the root attack and how it can be used to exploit the " +
            "system specifications.\"}} " +
            "# Respect: All words in the \"name\" of the root attack must be conjoined together. " +
            "# Respect: There must be no more than forty characters in the \"name\" of the root attack. " +
            "# Respect: For each word in the \"name\" of the root attack, its first letter must be capitalized. " +
            "# Respect: Include what the root attack is and how it can be used to exploit the system " +
            "specification in the same \"description\" key." +
            "# Respect: All words in the \"description\" key must be separated with spaces.";

    private static final String KNOWLEDGE_ON_JSON_FOR_ATTACKS = "When you are asked to identify the attack nodes " +
            "that an attacker needs to complete to successfully achieve the root attack, " +
            "return them as a JSON specification formatted as follows: " +
            "{\"attacknodes\": [{\"name\":  \"NameOfAttackNode\", \"description\": \"" +
            "The description of the attack node and how it brings an attacker closer to the root attack.\"} ...]} " +
            "# Respect: All words in the \"name\" of each attack node must be conjoined together. " +
            "# Respect: There must be no more than forty characters in the \"name\" of each attack node. " +
            "# Respect: For each word in each attack node's \"name\", its first letter must be capitalized. " +
            "# Respect: Include what the attack node is and how it is used by an attacker for getting closer " +
            "to the root attack in the same \"description\" key. " +
            "# Respect: All words in the \"description\" key must be separated with spaces.";

    private static final String KNOWLEDGE_ON_JSON_FOR_ATT_CONNS = "When you are asked to identify connections between the root " +
            "attack and the attack nodes, return them as a JSON specification formatted as follows: " +
            "{\"attackconnections\": [{\"parentattack\": \"NameOfRootAttack or NameOfAttackNode\", " +
            "\"connectiontype\": \"the connection type\", " +
            "\"childrenattacks\": [\"NameOfAttackNode\" ...]} ...]} " +
            "# Respect: Each attack connection must have at least two children attack nodes. " +
            "# Respect: The \"childrenattacks\" array must not contain the root attack. " +
            "# Note: There are four types of connections: \"OR\", \"XOR\", \"AND\", and \"SEQUENCE\". " +
            "# Note: An \"OR\" connection represents the scenario that among all of the children attacks, " +
            "an attacker only needs one of the children to proceed to the parent attack. " +
            "# Note: A \"XOR\" connection represents the scenario that among all of the children attacks, " +
            "an attacker only needs one and only one of the children to proceed to the parent attack. " +
            "# Note: An \"AND\" connection represents the scenario that among all of the children objects, " +
            "an attacker needs all children simultaneously to proceed to the parent attack. " +
            "# Note: A \"SEQUENCE\" connection represents the scenario that an attacker needs each of the children " +
            "objects sequentially to proceed to the parent attack. The first indexed child of a \"SEQUENCE\" " +
            "connection is the child that an attacker needs first while the last indexed child is the child that " +
            "an attacker needs last. " +
            "# Respect: The \"connectiontype\" must only be \"OR\", \"XOR\", \"AND\", or \"SEQUENCE\".";

    private static final String KNOWLEDGE_ON_JSON_FOR_MITIGATIONS = "When you are asked to identify the mitigations " +
            "that prevent an attacker from performing an attack node, " +
            "return them as a JSON specification formatted as follows: " +
            "{\"mitigations\": [{\"name\":  \"NameOfMitigation\", \"description\": \"" +
            "The description of the mitigation and how it prevents an attacker from completing " +
            "its associated attack node.\"} ...]} " +
            "# Respect: All words in the \"name\" of each mitigation must be conjoined together. " +
            "# Respect: There must be no more than forty characters in the \"name\" of each mitigation. " +
            "# Respect: For each word in each mitigation's \"name\", its first letter must be capitalized. " +
            "# Respect: Include what the mitigation is and how it is used to prevent an attacker from " +
            "completing its associated attack node in the same \"description\" key. " +
            "# Respect: If there are no mitigations that can be applied to the provided attack nodes, return the " +
            "following JSON: {\"mitigations\": \"No mitigations were able to be identified.\"} " +
            "# Respect: All words in the \"description\" key must be separated with spaces. ";

    private static final String KNOWLEDGE_ON_JSON_FOR_MITI_PAIRS = "When you are asked to identify the " +
            "mitigation that can be applied to an attack node to prevent " +
            "an attacker from performing the attack node, " +
            "return them as a JSON specification formatted as follows: " +
            "{\"mitigationpairings\": [{\"mitigation\":  \"NameOfMitigation\", \"attacknode\": \"NameOfAttackNode\"} ...]} " +
            "# Respect: The value of \"mitigation\" should be only one of the names of the given mitigations. " +
            "# Respect: The value of \"attacknode\" should be only one of the names of the given attack nodes. " +
            "# Respect: The value of \"attacknode\" should not be the name of the root attack.";

    private static final String[] KNOWLEDGE_STAGES = {
            KNOWLEDGE_ON_JSON_FOR_ROOT,
            KNOWLEDGE_ON_JSON_FOR_ATTACKS,
            KNOWLEDGE_ON_JSON_FOR_ATT_CONNS
//            KNOWLEDGE_ON_JSON_FOR_MITIGATIONS,
//            KNOWLEDGE_ON_JSON_FOR_MITI_PAIRS
    };

    private final String[] QUESTION_IDENTIFY_ATD = {"From the provided system specification and using the specified " +
            "JSON format, identify a possible root attack that an attacker could potentially exploit on the provided " +
            "system specification. Do respect the JSON format, and " +
            "provide only JSON (no explanation before or after).\n",

            "From the provided system specification and root attack and " +
                    "using the specified JSON format, identify the attack nodes that an attacker needs to " +
                    "complete for achieving the root attack. Do respect the JSON format, and " +
                    "provide only JSON (no explanation before or after).\n",

            "From the provided system specification, root attack, and attack nodes and " +
                    "using the specified JSON format, identify the connections that illustrate in what order an " +
                    "attacker needs to complete the attack nodes to achieve the root attack. " +
                    "Do respect the JSON format, and provide only JSON (no explanation before or after).\n",

            "From the provided system specification and attack nodes " +
                    "and using the specified JSON format, " +
                    "identify possible mitigations, if there are any, that could prevent an attacker from completing " +
                    "an attack node. Do respect the JSON format, and provide only JSON (no explanation before or after).\n",

            "From the provided system specification, attack nodes, and mitigations " +
                    "and using the specified JSON format, " +
                    "identify what mitigation can be applied to which attack node such that the mitigation prevents " +
                    "an attacker from performing the attack node. " +
                    "Do respect the JSON format, and provide only JSON (no explanation before or after).\n"
    };

    private String rootAttackData;
    private String attackNodeData;
    private String attConnections;
    private String mitigations;
    private String mitiPairs;
    private AttackTree at;
    private ArrayList<AttackTree> atList;

    public AIAttackPatternTree1(AIChatData _chatData) {
        super(_chatData);
        atList = new ArrayList<>();
    }

    public AttackTree getATDiagram() {
        return at;
    }

    public void internalRequest() {
        String previousRootAttacks = "";

        for (int i = 0; i < 5; i++) {
            at = new AttackTree("", null);
            int stage = 0;
            String lastQuestion = chatData.lastQuestion.trim();
            String json = "";
            String questionT = QUESTION_IDENTIFY_ATD[stage];

            initKnowledge();
            makeKnowledge(stage, lastQuestion, previousRootAttacks);

            boolean done = false;
            int cpt = 0;

            // actors, use cases and connections
            while (!done && cpt < 40) {
                cpt++;
                boolean ok = makeQuestion(questionT);

                if (!ok) {
                    done = true;
                    TraceManager.addDev("Make question failed");
                }

                ArrayList<String> errors = null;

                try {
                    TraceManager.addDev("\n\nMaking specification from " + chatData.lastAnswer + "\n\n");
                    json = extractJSON();

                    if (stage == 0) {
                        rootAttackData = "";

                        errors = new ArrayList<>();
                        rootAttackData = checkRootAttack(json, errors);
                        TraceManager.addDev("Identified root attack - " + rootAttackData);

                        if (rootAttackData.isEmpty()) {
                            errors.add("You must provide the root attack of " +
                                    "the given attack pattern and how it can be used to " +
                                    "exploit the provided system specification. Do respect the JSON format, and " +
                                    "provide only JSON (no explanation before or after).");
                        }
                    } else if (stage == 1) {
                        attackNodeData = "";

                        errors = new ArrayList<>();
                        attackNodeData = checkAttackNodes(json, errors);
                        TraceManager.addDev("Identified attack nodes: " + attackNodeData);

                        if (attackNodeData.isEmpty()) {
                            errors.add("You must provide the attack nodes showing how an " +
                                    "attacker uses these nodes to achieve the root attack. " +
                                    "Do respect the JSON format, and " +
                                    "provide only JSON (no explanation before or after).");
                        }
                    } else if (stage == 2) {
                        attConnections = "";

                        errors = new ArrayList<>();
                        attConnections = checkAttackConns(json, errors);
                        TraceManager.addDev("Identified attack connections: " + attConnections);

                        if (attConnections.isEmpty()) {
                            errors.add("You must provide the connections showing in what order " +
                                    "an attacker needs to complete the attack nodes to achieve the " +
                                    "root attack. Do respect the JSON format, and " +
                                    "provide only JSON (no explanation before or after).");
                        }
                    } else if (stage == 3) {
                        mitigations = "";

                        errors = new ArrayList<>();
                        mitigations = checkMitigations(json, errors);

                        TraceManager.addDev("Identified mitigations: " + mitigations);

                        if (mitigations.isEmpty()) {
                            errors.add("You must provide mitigations showing how they prevent " +
                                    "an attacker from performing an attack node. " +
                                    "Do respect the JSON format, and " +
                                    "provide only JSON (no explanation before or after).");
                        }
                    } else if (stage == 4) {
                        mitiPairs = "";

                        errors = new ArrayList<>();
                        mitiPairs = checkMitiPairs(json, errors);

                        TraceManager.addDev("Identified mitigation pairings: " + mitiPairs);

                        if (mitiPairs.isEmpty()) {
                            errors.add("You must associate the provided mitigations with a provided attack node " +
                                    "such that the mitigation prevents an attacker from performing the attack node. " +
                                    "Do respect the JSON format, and " +
                                    "provide only JSON (no explanation before or after).");
                        }
                    }
                } catch (org.json.JSONException e) {
                    TraceManager.addDev("Invalid JSON spec: " + extractJSON() + " because " + e.getMessage() + ": INJECTING ERROR");
                    errors = new ArrayList<>();
                    errors.add("There is an error in your JSON: " + e.getMessage() + ". Probably the JSON spec was incomplete. Do correct it. I need " +
                            "the full specification at once.");
                }

                if ((errors != null) && (!errors.isEmpty())) {
                    questionT = "Your answer was as follows: " + json + "\n\nYet, it was not correct because of the following errors:";
                    // Updating knowledge
                    for (String s : errors) {
                        questionT += "\n- " + s;
                    }

                    initKnowledge();

                    if (stage == 0) {
                        makeKnowledge(stage, lastQuestion, previousRootAttacks);
                    }
                    else {
                        makeKnowledge(stage, lastQuestion, "");
                    }
                } else {
                    stage++;

                    if (stage == KNOWLEDGE_STAGES.length) {
                        done = true;
                    } else {
                        initKnowledge();
                        makeKnowledge(stage, lastQuestion, "");
                        questionT = QUESTION_IDENTIFY_ATD[stage];

                        if (stage == 1) {
                            questionT += "\nThe root attack data is in the following JSON:\n" + rootAttackData.trim() + "\n";
                        } else if (stage == 2) {
                            questionT += "\nThe root attack data is in the following JSON:\n" + rootAttackData.trim() + "\n";
                            questionT += "\nThe data of the attack nodes are in the following JSON:\n" +
                                    attackNodeData.trim() + "\n";
                        } else if (stage == 3) {
                            questionT += "\nThe data of the attack nodes are in the following JSON:\n" + attackNodeData.trim();
                        } else if (stage == 4) {
                            if (mitigations.toLowerCase().contains("no mitigations were able to be identified")) {
                                done = true;
                            } else {
                                questionT += "\nThe data of the attack nodes are in the following JSON:\n" + attackNodeData.trim();
                                questionT += "\nThe data of the mitigations are in the following JSON:\n" + mitigations.trim();
                            }
                        }
                    }
                }

                waitIfConditionTrue(!done && cpt < 20);
                cpt++;
            }

            TraceManager.addDev("Reached end of AIAttackPatternTree internal request cpt=" + cpt);
            atList.add(at);
            previousRootAttacks = previousRootAttacks + rootAttackData + " ";
        }
    }

    public Object applyAnswer(Object input) { return atList; }

    private void initKnowledge() {
        chatData.aiinterface.clearKnowledge();
    }

    private void makeKnowledge(int stage, String _spec, String previousRootAttacks) {
        String [] know = KNOWLEDGE_STAGES[stage].split("#");
        for(String s: know) {
            TraceManager.addDev("\nKnowledge added: " + s);
            chatData.aiinterface.addKnowledge(s, "ok");
        }

        if (_spec != null) {
            TraceManager.addDev("\nKnowledge added: " + _spec);
            chatData.aiinterface.addKnowledge("The system specification is: " + _spec, "ok");
        }

        if (!previousRootAttacks.equals("")) {
            TraceManager.addDev("\nKnowledge added: " + previousRootAttacks);
            chatData.aiinterface.addKnowledge(
                    "Identify a root attack other than the root attacks in the following JSONs: " + previousRootAttacks,
                    "ok");

        }
    }

    private String checkRootAttack(String _spec, Collection<String> _errors) throws org.json.JSONException {
        if (_spec == null) {
            _errors.add("No \"rootattack\" object in json");

            return "";
        }

        int indexStart = _spec.indexOf('{');
        int indexStop = _spec.lastIndexOf('}');

        if ((indexStart == -1) || (indexStop == -1) || (indexStart > indexStop)) {
            _errors.add("Invalid JSON object (start or stop)");

            return "";
        }

        String json = _spec.substring(indexStart, indexStop + 1);

        JSONObject mainObject = new JSONObject(json);
        JSONObject rootAttackJSON = mainObject.getJSONObject("rootattack");

        if (rootAttackJSON == null) {
            TraceManager.addDev("No \"rootattack\" array in json");
            _errors.add("No \"rootattack\" array in json");

            return "";
        }
        else {
            _errors.addAll(checkAttackSyntax(rootAttackJSON));
        }

        if (_errors.isEmpty()) {
            Attack rootAttack = new Attack("", null);
            rootAttack.setRoot(true);
            rootAttack.setName(rootAttackJSON.getString("name"));
            rootAttack.setDescription(rootAttackJSON.getString("description"));
            at.addAttack(rootAttack);

            return _spec;
        }
        else {
            return "";
        }
    }

    private String checkAttackNodes(String _spec, Collection<String> _errors) throws org.json.JSONException {
        if (_spec == null) {
            _errors.add("No \"attacknodes\" array in json");

            return "";
        }

        int indexStart = _spec.indexOf('{');
        int indexStop = _spec.lastIndexOf('}');

        if ((indexStart == -1) || (indexStop == -1) || (indexStart > indexStop)) {
            _errors.add("Invalid JSON object (start or stop)");

            return "";
        }

        String json = _spec.substring(indexStart, indexStop + 1);

        JSONObject mainObject = new JSONObject(json);
        JSONArray attackNodesJSON = mainObject.getJSONArray("attacknodes");

        if (attackNodesJSON == null) {
            TraceManager.addDev("No \"attacknodes\" array in json");
            _errors.add("No \"attacknodes\" array in json");

            return "";
        }
        else {
            int i = 0;

            while (i < attackNodesJSON.length() && _errors.isEmpty())
            {
                _errors.addAll(checkAttackSyntax(attackNodesJSON.getJSONObject(i)));
                i++;
            }
        }

        if (_errors.isEmpty()) {
            for (int i = 0; i < attackNodesJSON.length(); i++) {
                JSONObject attackNode = attackNodesJSON.getJSONObject(i);

                Attack attack = new Attack("", null);
                attack.setName(attackNode.getString("name"));
                attack.setDescription(attackNode.getString("description"));
                at.addAttack(attack);
            }

            return _spec;
        }
        else {
            return "";
        }
    }

    private ArrayList<String> checkAttackSyntax(JSONObject attack) {
        ArrayList<String> errors = new ArrayList<>();

        if (attack.length() != 2) {
            errors.add("The attack object should only contain \"name\" and \"description\"");

            return errors;
        }

        String name = attack.getString("name");
        String description = attack.getString("description");

        if (name == null) {
            errors.add("Attack has no name");
        }
        else if (!name.matches("[a-zA-Z0-9]+")) {
            errors.add(name + " must only contain alphanumeric characters");
        }
        else if (name.length() > 40) {
            errors.add(name + " must only contain forty characters max");
        }
        else if (description == null) {
            errors.add("Attack has no description");
        }
        else if (!description.contains(" ")) {
            errors.add("The words in \"description\" must be separated with spaces.");
        }

        return errors;
    }

    private String checkAttackConns(String _spec, Collection<String> _errors) throws org.json.JSONException {
        if (_spec == null) {
            _errors.add("No \"attackconnections\" array in json");

            return "";
        }

        int indexStart = _spec.indexOf('{');
        int indexStop = _spec.lastIndexOf('}');

        if ((indexStart == -1) || (indexStop == -1) || (indexStart > indexStop)) {
            _errors.add("Invalid JSON object (start or stop)");

            return "";
        }

        String json = _spec.substring(indexStart, indexStop + 1);

        JSONObject mainObject = new JSONObject(json);
        JSONArray attackConnsJSON = mainObject.getJSONArray("attackconnections");

        if (attackConnsJSON == null) {
            TraceManager.addDev("No \"attackconnections\" array in json");
            _errors.add("No \"attackconnections\" array in json");

            return "";
        }
        else {
            int i = 0;

            while (i < attackConnsJSON.length() && _errors.isEmpty())
            {
                _errors.addAll(checkAttConnSyntax(attackConnsJSON.getJSONObject(i)));
                i++;
            }
        }

        if (_errors.isEmpty()) {
            AttackNode connection = null;

            for (int i = 0; i < attackConnsJSON.length(); i++) {
                JSONObject attackConn = attackConnsJSON.getJSONObject(i);
                String parentAttackName = attackConn.getString("parentattack");
                String connectionType = attackConn.getString("connectiontype");
                JSONArray childrenAttackNames = attackConn.getJSONArray("childrenattacks");
                List<Attack> attackList = at.getAttacks();

                switch(connectionType) {
                    case "OR":
                        connection = new ORNode("", null);
                        break;
                    case "XOR":
                        connection = new XORNode("", null);
                        break;
                    case "AND":
                        connection = new ANDNode("", null);
                        break;
                    default:
                        connection = new SequenceNode("", null);
                        break;
                }

                Attack parentAttack = findAttack(parentAttackName, attackList, true);
                connection.setResultingAttack(parentAttack);
                int attackValue = 1;

                for (Object childAttackName : childrenAttackNames) {
                    String childAttackNameString = (String) childAttackName;
                    Attack childAttack = findAttack(childAttackNameString, attackList, false);

                    connection.addInputAttack(childAttack, attackValue);

                    attackValue++;
                }

                at.addNode(connection);
            }

            return _spec;
        }
        else {
            return "";
        }
    }

    private ArrayList<String> checkAttConnSyntax(JSONObject attackConn) {
        ArrayList<String> errors = new ArrayList<>();

        if (attackConn.length() != 3) {
            errors.add("The attack connection object should only contain " +
                    "\"parentattack\", \"connectiontype\", and \"childrenattacks\".");

            return errors;
        }

        String parentAttack = attackConn.getString("parentattack");
        String connectionType = attackConn.getString("connectiontype");
        JSONArray childrenAttacks = attackConn.getJSONArray("childrenattacks");

        if (parentAttack == null) {
            errors.add("Connection has no parentattack");
            return errors;
        }

        Attack foundParentAttack = findAttack(parentAttack, at.getAttacks(), true);

        if (foundParentAttack == null) {
            errors.add(parentAttack + " is not the name of any provided root attack " +
                    "or attack node. Ensure that \"parentattack\" is the name of either the provided root" +
                    "attack or one of the attack nodes.");
        }

        if (connectionType == null) {
            errors.add("Connection has no connectiontype");
        }
        else if (!connectionType.equals("OR") &&
                !connectionType.equals("XOR") &&
                !connectionType.equals("AND") &&
                !connectionType.equals("SEQUENCE")) {
            errors.add("Ensure that connectiontype is one of the following types: \"OR\", \"XOR\", \"AND\", " +
                    "or \"SEQUENCE\"");
        }

        if (childrenAttacks == null) {
            errors.add("Connection has no childrenattacks");
            return errors;
        }
//        else if (childrenAttacks.length() <= 1) {
//            errors.add("Connection has only one child attack. Please ensure that all connections have at least two " +
//                    "children attacks.");
//            return errors;
//        }

        for (Object childAttack : childrenAttacks) {
            String childAttackString = (String) childAttack;
            Attack foundChildAttack = findAttack(childAttackString, at.getAttacks(), false);

            if (foundChildAttack == null) {
                errors.add(childAttackString + " is not the name of any provided attack node. " +
                        "Ensure that all values in the \"childrenattacks\" array are one of the names in " +
                        "the \"attacknodes\" JSON array.");
            }
        }

        return errors;
    }

    private Attack findAttack(String attackNodeName, List<Attack> attackNodeList, boolean checkRoot) {
        Attack attack = null;
        int i;

        if (!checkRoot) {
            i = 1;
        }
        else {
            i = 0;
        }

        while (i < attackNodeList.size()) {
            attack = attackNodeList.get(i);

            if (attackNodeName.equals(attack.getName())) {
                return attack;
            }

            i++;
        }

        return null;
    }

    private String checkMitigations(String _spec, Collection<String> _errors)
            throws org.json.JSONException {
        String mitigationList = "";

        if (_spec == null) {
            _errors.add("No \"mitigations\" array in json");

            return "";
        }

        int indexStart = _spec.indexOf('{');
        int indexStop = _spec.lastIndexOf('}');

        if ((indexStart == -1) || (indexStop == -1) || (indexStart > indexStop)) {
            _errors.add("Invalid JSON object (start or stop)");

            return "";
        }

        String json = _spec.substring(indexStart, indexStop + 1);

        JSONObject mainObject = new JSONObject(json);
        Object mitigations = mainObject.get("mitigations");

        JSONArray mitigationsJSON = (JSONArray) mitigations;

        if (mitigationsJSON == null) {
            TraceManager.addDev("No \"mitigations\" array in json");
            _errors.add("No \"mitigations\" array in json");

            return "";
        }
        else {
            int i = 0;

            while (i < mitigationsJSON.length() && _errors.isEmpty())
            {
                _errors.addAll(checkMitigationSyntax(mitigationsJSON.getJSONObject(i)));
                i++;
            }
        }

        if (_errors.isEmpty()) {
//            for (int i = 0; i < mitigationsJSON.length(); i++) {
//                JSONObject attackStep = mitigationsJSON.getJSONObject(i);
//
//                Attack attack = new Attack("", null);
//                attack.setName(attackStep.getString("name"));
//                attack.setDescription(attackStep.getString("description"));
//                at.addAttack(attack);
//
//                mitigations += attack.toString();
//            }

            return _spec;
        }
        else {
            return "";
        }
    }

    private ArrayList<String> checkMitigationSyntax(JSONObject mitigation) {
        ArrayList<String> errors = new ArrayList<>();

        if (mitigation.length() != 2) {
            errors.add("The mitigation object should only contain \"name\" and \"description\"");

            return errors;
        }

        String name = mitigation.getString("name");
        String description = mitigation.getString("description");
//        String attackNodeName = mitigation.getString("attacknode");

        if (name == null) {
            errors.add("Mitigation has no name");
        }
        else if (!name.matches("[a-zA-Z0-9]+")) {
            errors.add(name + " must only contain alphanumeric characters");
        }
        else if (name.length() > 40) {
            errors.add(name + " must only contain forty characters max");
        }
        else if (description == null) {
            errors.add("Mitigation has no description");
        }
        else if (!description.contains(" ")) {
            errors.add("The words in \"description\" must be separated with spaces.");
        }
//        else if (attackNodeName == null) {
//            errors.add("Mitigation has no associated attack node");
//        }
//        else if (attackNodeName.equals(at.getAttacks().get(0).getName())) {
//            errors.add(attackNodeName + " is the name of the root attack. " +
//                    "Do not associate mitigations with the root attack");
//        }

//        ArrayList<Attack> attackNodeList = at.getAttacks();
//        ArrayList<String> differentAttackNames = getAllDiffAttackNodeNames(attackNodeName, attackNodeList);
//        if (differentAttackNames.size() == at.getAttacks().size() - 1) {
//            StringBuilder builder = new StringBuilder(
//                    attackNodeName + " in \"attacknode\" is not the name of any of the provided attack nodes. " +
//                    "The value of \"attacknode\" can only be one of the following names: ");
//
//            for (int i = 0; i < differentAttackNames.size(); i++) {
//                if (i != differentAttackNames.size() - 1) {
//                    builder.append("\"").append(differentAttackNames.get(i)).append("\"").append(", ");
//                }
//                else {
//                    builder.append("or ").append("\"").append(differentAttackNames.get(i)).append("\"");
//                }
//            }

//            errors.add(builder.toString());
//        }

        return errors;
    }

    private String checkMitiPairs(String _spec, Collection<String> _errors)
            throws org.json.JSONException {
        String mitigationList = "";

        if (_spec == null) {
            _errors.add("No \"mitigationpairings\" array in json");

            return "";
        }

        int indexStart = _spec.indexOf('{');
        int indexStop = _spec.lastIndexOf('}');

        if ((indexStart == -1) || (indexStop == -1) || (indexStart > indexStop)) {
            _errors.add("Invalid JSON object (start or stop)");

            return "";
        }

        String json = _spec.substring(indexStart, indexStop + 1);

        JSONObject mainObject = new JSONObject(json);
        JSONArray mitigationPairingsJSON = mainObject.getJSONArray("mitigationpairings");

        if (mitigationPairingsJSON == null) {
            TraceManager.addDev("No \"mitigationpairings\" array in json");
            _errors.add("No \"mitigationpairings\" array in json");

            return "";
        }
        else {
            int i = 0;

            while (i < mitigationPairingsJSON.length() && _errors.isEmpty())
            {
                _errors.addAll(checkMitiPairingSyntax(mitigationPairingsJSON.getJSONObject(i)));
                i++;
            }
        }

        if (_errors.isEmpty()) {
//            for (int i = 0; i < mitigationsJSON.length(); i++) {
//                JSONObject attackStep = mitigationsJSON.getJSONObject(i);
//
//                Attack attack = new Attack("", null);
//                attack.setName(attackStep.getString("name"));
//                attack.setDescription(attackStep.getString("description"));
//                at.addAttack(attack);
//
//                mitigations += attack.toString();
//            }

            return _spec;
        }
        else {
            return "";
        }
    }

    private ArrayList<String> checkMitiPairingSyntax(JSONObject mitigation) {
        ArrayList<String> errors = new ArrayList<>();

        if (mitigation.length() != 2) {
            errors.add("The mitigation pairing object should only contain \"mitigation\" and \"attacknode\"");

            return errors;
        }

        String mitigationName = mitigation.getString("mitigation");
        String attackNodeName = mitigation.getString("attacknode");

        if (mitigationName == null) {
            errors.add("Pairing has no mitigation");
        }
        else if (attackNodeName == null) {
            errors.add("Pairing has no attack node");
        }

        Attack foundAttackNode = findAttack(attackNodeName, at.getAttacks(), false);

        if (foundAttackNode == null) {
            errors.add(attackNodeName + " is not the name of any provided attack node. " +
                    "Ensure that the value of \"attacknode\" is the name of an attack node.");
        }

        return errors;
    }
}
